import pathlib
import re
import yaml
import pandas as pd
from collections import namedtuple

from ..utilities import get_configuration


def _read_yaml_config(config_path):
    with open(config_path, 'r') as f:
        config = yaml.safe_load(f)
    return config


def _read_ini_config(config_path):
    return get_configuration(config_path)


def read_mapping_config(cwd: str = '.'):
    tried = []
    yaml_path = None
    for name in ['config', 'mapping_config']:
        for config_dir in [cwd, f'{cwd}/..']:
            for suffix in ['yaml', 'yml']:
                path = f'{config_dir}/{name}.{suffix}'
                tried.append(path)
                if pathlib.Path(path).exists():
                    yaml_path = path

    ini_path = None
    for name in ['config', 'mapping_config']:
        for config_dir in [cwd, f'{cwd}/..']:
            path = f'{config_dir}/{name}.ini'
            tried.append(path)
            if pathlib.Path(path).exists():
                ini_path = path

    if yaml_path is not None:
        config_flat_dict = _read_yaml_config(yaml_path)
    elif ini_path is not None:
        config_flat_dict = _read_ini_config(ini_path)
    else:
        tried_str = '\n'.join(tried)
        raise FileNotFoundError(f'No mapping config file found, '
                                f'tried to search these paths but none exist: '
                                f'{tried_str}')

    config = namedtuple('config', config_flat_dict.keys())(*config_flat_dict.values())
    return config, config_flat_dict


def validate_cwd_fastq_paths(cwd: str = '.'):
    """
    Validate fastq paths in the fastq subdirectory of cwd.
    Parameters
    ----------
    cwd :
        Path of the current working directory.

    Returns
    -------
    fastq_table : pandas.DataFrame
    """
    # get all fastq file paths
    fastq_paths = [p
                   for p in pathlib.Path(f'{cwd}/fastq/').glob('*.[fq.gz][fastq.gz]')
                   if 'trim' not in p.name]

    # parse cell id and match fastq pairs
    fastq_pattern = re.compile(r'(?P<cell_id>.+)(-|_)(?P<read_type>(R1|R2|r1|r2)).(fastq|fq)(.gz)*')
    fastq_records = {}
    for p in fastq_paths:
        match = fastq_pattern.match(p.name)
        if match is None:
            # print(f'WARNING: {p} has FASTQ file path suffix, but do not match '
            #       f'expected file name pattern {fastq_pattern}')
            pass
        else:
            cell_id = match.group('cell_id')
            read_type = match.group('read_type')
            fastq_records[cell_id, read_type.upper()] = str(p)

    if len(fastq_records) == 0:
        raise ValueError('No fastq files found in fastq folder, '
                         'or no fastq files match expected file name pattern')

    fastq_table = pd.Series(fastq_records).unstack()
    if 'R1' not in fastq_table.columns or 'R2' not in fastq_table.columns:
        raise ValueError('No R1 or R2 fastq files found')
    fastq_table = fastq_table[['R1', 'R2']].copy()

    # raise error if fastq file not paired
    missing_file = fastq_table.isna().sum(axis=1) > 0
    if missing_file.sum() > 0:
        for cell in missing_file[missing_file].index:
            print(f'{cell} missing R1 or R2 FASTQ file.')
        raise FileNotFoundError(f'FASTQ files in {pathlib.Path(f"{cwd}/fastq/").absolute()} is not all paired.')
    return fastq_table



# "DELETE" means this column will be removed from the output file
# "" means the name will not be changed
# "CertainMetricNames" means the name will be changed to "CertainMetricNames"


COL_NAMES = {
    ('cell_parser_cutadapt_trim_stats', 'status'): 'DELETE',
    ('cell_parser_cutadapt_trim_stats', 'in_reads'): 'InputReadPairs',
    ('cell_parser_cutadapt_trim_stats', 'in_bp'): 'InputReadPairsBP',
    ('cell_parser_cutadapt_trim_stats', 'too_short'): 'DELETE',
    ('cell_parser_cutadapt_trim_stats', 'too_long'): 'DELETE',
    ('cell_parser_cutadapt_trim_stats', 'too_many_n'): 'DELETE',
    ('cell_parser_cutadapt_trim_stats', 'out_reads'): 'TrimmedReadPairs',
    ('cell_parser_cutadapt_trim_stats', 'w/adapters'): 'R1WithAdapters',
    ('cell_parser_cutadapt_trim_stats', 'qualtrim_bp'): 'R1QualTrimBP',
    ('cell_parser_cutadapt_trim_stats', 'out_bp'): 'R1TrimmedReadsBP',
    ('cell_parser_cutadapt_trim_stats', 'w/adapters2'): 'R2WithAdapters',
    ('cell_parser_cutadapt_trim_stats', 'qualtrim2_bp'): 'R2QualTrimBP',
    ('cell_parser_cutadapt_trim_stats', 'out2_bp'): 'R2TrimmedReadsBP',
    ('cell_parser_hisat_summary', 'ReadPairsMappedInPE'): 'DELETE',
    ('cell_parser_hisat_summary', 'PEUnmappableReadPairs'): 'DELETE',
    ('cell_parser_hisat_summary', 'PEUniqueMappedReadPairs'): 'DELETE',
    ('cell_parser_hisat_summary', 'PEMultiMappedReadPairs'): 'DELETE',
    ('cell_parser_hisat_summary', 'PEDiscordantlyUniqueMappedReadPairs'): 'DELETE',
    ('cell_parser_hisat_summary', 'ReadsMappedInSE'): 'DELETE',
    ('cell_parser_hisat_summary', 'SEUnmappableReads'): 'DELETE',
    ('cell_parser_hisat_summary', 'SEUniqueMappedReads'): 'DELETE',
    ('cell_parser_hisat_summary', 'SEMultiMappedReads'): 'DELETE',
    ('cell_parser_hisat_summary', 'UniqueMappedReads'): 'UniqueMappedReads',
    ('cell_parser_hisat_summary', 'MultiMappedReads'): 'MultiMappedReads',
    ('cell_parser_hisat_summary', 'UniqueMappingRate'): 'UniqueMappingRate',
    ('cell_parser_hisat_summary', 'MultiMappingRate'): 'MultiMappingRate',
    ('cell_parser_hisat_summary', 'OverallMappingRate'): 'OverallMappingRate',
    ('cell_parser_picard_dedup_stat', 'LIBRARY'): 'DELETE',
    ('cell_parser_picard_dedup_stat', 'UNPAIRED_READS_EXAMINED'): 'DELETE',
    ('cell_parser_picard_dedup_stat', 'READ_PAIRS_EXAMINED'): 'DELETE',
    ('cell_parser_picard_dedup_stat', 'SECONDARY_OR_SUPPLEMENTARY_RDS'): 'DELETE',
    ('cell_parser_picard_dedup_stat', 'UNMAPPED_READS'): 'DELETE',
    ('cell_parser_picard_dedup_stat', 'UNPAIRED_READ_DUPLICATES'): 'DELETE',
    ('cell_parser_picard_dedup_stat', 'READ_PAIR_DUPLICATES'): 'DELETE',
    ('cell_parser_picard_dedup_stat', 'READ_PAIR_OPTICAL_DUPLICATES'): 'DELETE',
    ('cell_parser_picard_dedup_stat', 'PERCENT_DUPLICATION'): 'DELETE',
    ('cell_parser_picard_dedup_stat', 'ESTIMATED_LIBRARY_SIZE'): 'DELETE',
    ('cell_parser_picard_dedup_stat', 'FinalReads'): '',
    ('cell_parser_picard_dedup_stat', 'DuplicatedReads'): '',
    ('cell_parser_picard_dedup_stat', 'PCRDuplicationRate'): '',
    ('cell_parser_feature_count_summary', 'Assigned'): 'AssignedRNAReads',
    ('cell_parser_feature_count_summary', 'Unassigned_Unmapped'): 'DELETE',
    ('cell_parser_feature_count_summary', 'Unassigned_Read_Type'): 'DELETE',
    ('cell_parser_feature_count_summary', 'Unassigned_Singleton'): 'DELETE',
    ('cell_parser_feature_count_summary', 'Unassigned_MappingQuality'): 'DELETE',
    ('cell_parser_feature_count_summary', 'Unassigned_Chimera'): 'DELETE',
    ('cell_parser_feature_count_summary', 'Unassigned_FragmentLength'): 'DELETE',
    ('cell_parser_feature_count_summary', 'Unassigned_Duplicate'): 'DELETE',
    ('cell_parser_feature_count_summary', 'Unassigned_MultiMapping'): 'DELETE',
    ('cell_parser_feature_count_summary', 'Unassigned_Secondary'): 'DELETE',
    ('cell_parser_feature_count_summary', 'Unassigned_NonSplit'): 'DELETE',
    ('cell_parser_feature_count_summary', 'Unassigned_NoFeatures'): 'DELETE',
    ('cell_parser_feature_count_summary', 'Unassigned_Overlapping_Length'): 'DELETE',
    ('cell_parser_feature_count_summary', 'Unassigned_Ambiguity'): 'DELETE',
    ('cell_parser_feature_count_summary', 'Unassigned_Total'): 'UnassignedRNAReads',
    ('cell_parser_feature_count_summary', 'AssignedRNAReadsRate'): 'AssignedRNAReadsRate',
}



import pandas as pd


def _parse_cell_id_v1(cell_id):
    plate1, plate2, pcr_index, random_index = cell_id.split('-')
    if random_index.upper() in {'AD001', 'AD002', 'AD004', 'AD006'}:
        plate = plate1
    else:
        plate = plate2
    # 96 pos
    col96 = int(pcr_index[1:]) - 1
    row96 = ord(pcr_index[0]) - 65  # convert A-H to 0-8
    # 384 pos
    ad_index_384_dict = {
        'AD001': (0, 0),
        'AD002': (0, 1),
        'AD004': (1, 0),
        'AD006': (1, 1),
        'AD007': (0, 0),
        'AD008': (0, 1),
        'AD010': (1, 0),
        'AD012': (1, 1)
    }
    col384 = 2 * col96 + ad_index_384_dict[random_index][0]
    row384 = 2 * row96 + ad_index_384_dict[random_index][1]
    record = pd.Series({
        'Plate': plate,
        'PCRIndex': pcr_index,
        'RandomIndex': random_index,
        'Col384': col384,
        'Row384': row384
    })
    return record


def _parse_cell_id_v2(cell_id):
    plate, multiplex_group, pcr_index, random_index = cell_id.split('-')
    # 384 pos
    col384 = int(random_index[1:]) - 1
    row384 = ord(random_index[0]) - 65  # convert A-P to 0-23
    record = pd.Series({
        'Plate': plate,
        'PCRIndex': pcr_index,
        'MultiplexGroup': multiplex_group,
        'RandomIndex': random_index,
        'Col384': col384,
        'Row384': row384
    })
    return record


def get_plate_info(cell_ids, barcode_version):
    if barcode_version == 'V1':
        func = _parse_cell_id_v1
    else:
        func = _parse_cell_id_v2
    try:
        plate_info = pd.DataFrame([func(cell_id) for cell_id in cell_ids],
                                  index=cell_ids)
    except Exception:
        print('Errors occur during parsing the plate info, this happens '
              'when the input FASTQ file name is not generated by yap. '
              'The `yap summary` also can not generate html report due to missing the plate info. '
              'In this case, you need to add the plateinfo by yourself in order to make the plate view plots. '
              'These information is not necessary for following analysis though.')
        plate_info = pd.DataFrame([], index=cell_ids)
    return plate_info



import pandas as pd
import pathlib
import cemba_data

PACKAGE_DIR = pathlib.Path(cemba_data.__path__[0])


def prepare_mc_bulk(allc_table,
                    output_dir,
                    chrom_size_path,
                    mch_context='CHN',
                    mcg_context='CGN',
                    bigwig_mch_bin_size=50,
                    bigwig_mcg_bin_size=1,
                    cpu_per_job=12,
                    total_cpu=60):
    """
    Prepare the snakefile for merging single-cell ALLC files into pseudo-bulk

    Parameters
    ----------
    allc_table
        Path of the allc table. The allc table is a two column tsv file.
        The first columns is the absolute ALLC file paths;
        the second column is the group name of each file.
    output_dir
        Path of the output directory, will be created if not exist.
    mch_context
        mCH contexts for generating the bigwig tracks
    mcg_context
        mCG contexts for generating the bigwig tracks and merge strand
    chrom_size_path
        Path of the chromosome size file path
    bigwig_mch_bin_size
        Bin size used to generate mCH bigwig
    bigwig_mcg_bin_size
        Bin size used to generate mCG bigwig
    cpu_per_job
        Number of CPUs to use in individual merge-allc job
    total_cpu
        Number of CPUs to use in total

    Returns
    -------

    """
    snakemake_template_path = PACKAGE_DIR / 'bulk/Snakefile_template/mc_bulk.Snakefile'
    output_dir = pathlib.Path(output_dir).absolute()
    output_dir.mkdir(exist_ok=True)

    merge_allc_cpu = int(cpu_per_job / 1.1)
    total_mem_mb = cpu_per_job * 5000

    # prepare ALLC path dict
    # allc_path to group
    if str(allc_table).endswith('csv'):
        allc_path = pd.read_csv(allc_table, index_col=0, header=None, squeeze=True)
    else:
        allc_path = pd.read_csv(allc_table, sep='\t', index_col=0, header=None, squeeze=True)
    file_not_exist = allc_path[allc_path.index.map(lambda i: not pathlib.Path(i).exists())]
    if file_not_exist.size != 0:
        path_str = "\n".join(file_not_exist.index.tolist())
        raise FileNotFoundError(f'{file_not_exist.size} files do not exist:'
                                f'\n{path_str}')
    allc_dict = {group: paths.index.tolist() for group, paths in allc_path.groupby(allc_path)}

    # Prepare Snakefile
    snakemake_cmds = []
    for group, paths in allc_dict.items():
        # each group has a separate snakemake file
        group_dir = output_dir / group
        group_dir.mkdir(exist_ok=True)
        allc_list_path = group_dir / f'{group}.allc_paths.txt'
        with open(allc_list_path, 'w') as f:
            f.write('\n'.join(paths))
        snakemake_parameters = f"""
merge_allc_cpu = {merge_allc_cpu}
mch_context = '{mch_context}'
mcg_context = '{mcg_context}'
bigwig_mch_bin_size = {bigwig_mch_bin_size}
bigwig_mcg_bin_size = {bigwig_mcg_bin_size}
chrom_size_path = '{chrom_size_path}'
group = '{group}'

"""
        with open(snakemake_template_path) as f:
            snakemake_template = f.read()
        snakemake_str = snakemake_parameters + snakemake_template
        with open(group_dir / f'Snakefile', 'w') as f:
            f.write(snakemake_str)
        snakemake_cmd = f'snakemake ' \
                        f'-d {group_dir.absolute()} ' \
                        f'--snakefile {group_dir.absolute()}/Snakefile ' \
                        f'-j {cpu_per_job} ' \
                        f'--default-resources mem_mb=100 ' \
                        f'--resources mem_mb={total_mem_mb} ' \
                        f'--rerun-incomplete'
        snakemake_cmds.append(snakemake_cmd)

    qsub_dir = output_dir / 'qsub'
    qsub_dir.mkdir(exist_ok=True)
    with open(qsub_dir / 'snakemake_cmds.txt', 'w') as f:
        f.write('\n'.join(snakemake_cmds))
    with open(qsub_dir / 'qsub.sh', 'w') as f:
        qsub_str = f"""
yap qsub \
--command_file_path {qsub_dir / 'snakemake_cmds.txt'} \
--working_dir {qsub_dir} \
--project_name merge \
--total_cpu {total_cpu} \
--qsub_global_parms "-pe smp={cpu_per_job};-l h_vmem=5G"
"""
        f.write(qsub_str)
        print(f'Execute this command to start pipeline:\nnohup sh {qsub_dir / "qsub.sh"} &')
    return



from setuptools import setup, find_packages

setup(
    name='cemba-data',
    use_scm_version=True,
    setup_requires=['setuptools_scm'],
    author='Hanqing Liu',
    author_email='hanliu@salk.edu',
    description='Pipelines for single nucleus methylome and multi-omic dataset.',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/lhqing/cemba_data',
    license='MIT',
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
    ],
    packages=find_packages(exclude=('doc',)),
    include_package_data=True,
    package_data={
        '': ['*.txt', '*.tsv', '*.csv', '*.fa', '*Snakefile', '*ipynb']
    },
    install_requires=['pandas>=1.0',
                      'numpy',
                      'seaborn',
                      'matplotlib',
                      'papermill',
                      'dnaio',
                      'pysam'],
    entry_points={
        'console_scripts': ['yap=cemba_data.__main__:main',
                            'yap-internal=cemba_data._yap_internal_cli_:internal_main'],
    }
)



import pysam
import pathlib
import cemba_data
import subprocess
from ..utilities import get_configuration


def separate_unique_and_multi_align_reads(in_bam_path,
                                          out_unique_path,
                                          out_multi_path,
                                          out_unmappable_path=None,
                                          mapq_cutoff=10,
                                          qlen_cutoff=30):
    # TODO: make sure how to separate multi-align reads? or reads map to repeat regions in the genome?
    # TODO right now, we are just using mapq == 1 as multi-align reads, but this might not be right

    with pysam.AlignmentFile(in_bam_path, index_filename=None) as bam:
        header = bam.header
        with pysam.AlignmentFile(out_unique_path, header=header, mode='wb') as unique_bam, \
                pysam.AlignmentFile(out_multi_path, header=header, mode='wb') as multi_bam:
            if out_unmappable_path is not None:
                unmappable_bam = pysam.AlignmentFile(out_unmappable_path, header=header, mode='wb')
            else:
                unmappable_bam = None

            for read in bam:
                # skip reads that are too short
                if read.qlen < qlen_cutoff:
                    continue

                if read.mapq > mapq_cutoff:
                    unique_bam.write(read)
                elif read.mapq > 0:
                    multi_bam.write(read)
                else:
                    # unmappable reads
                    if unmappable_bam is not None:
                        unmappable_bam.write(read)

            if unmappable_bam is not None:
                unmappable_bam.close()
    return


def convert_hisat_bam_strandness(in_bam_path, out_bam_path):
    with pysam.AlignmentFile(in_bam_path) as in_bam, \
            pysam.AlignmentFile(out_bam_path, header=in_bam.header, mode='wb') as out_bam:
        for read in in_bam:
            if read.get_tag('YZ') == '+':
                read.is_forward = True
                if read.is_paired:
                    read.mate_is_forward = True
            else:
                read.is_forward = False
                if read.is_paired:
                    read.mate_is_forward = False
            out_bam.write(read)
    return


def make_snakefile_hisat3n(output_dir):
    output_dir = pathlib.Path(output_dir)

    mapping_config_name = list(output_dir.glob('mapping_config.*'))[0].name

    config = get_configuration(output_dir / mapping_config_name)
    try:
        mode = config['mode']
    except KeyError:
        raise KeyError('mode not found in the config file.')

    skip_dirs = ['stats', 'snakemake', 'scool']
    mapping_job_dirs = [p for p in output_dir.glob('*') if p.is_dir() and (p.name not in skip_dirs)]

    snakemake_dir = output_dir / 'snakemake'
    snakemake_dir.mkdir(exist_ok=True)
    stats_dir = output_dir / 'stats'
    stats_dir.mkdir(exist_ok=True)

    package_dir = cemba_data.__path__[0]
    snakefile_path = f'{package_dir}/hisat3n/snakefile/{mode.lower()}.Snakefile'
    if not pathlib.Path(snakefile_path).exists():
        print('Possible snakefile templates:')
        for p in pathlib.Path(f'{package_dir}/hisat3n/snakefile/').glob('Snakefile.*'):
            print(p)
        raise ValueError(f'Mode {mode} not supported, because Snakefile {snakefile_path} not found.')

    for p in mapping_job_dirs:
        subprocess.run(['cp', f'{output_dir}/{mapping_config_name}', f'{p}/{mapping_config_name}'], check=True)
        subprocess.run(['cp', snakefile_path, f'{p}/Snakefile'], check=True)

    # leave a flag to indicate using hisat-3n pipeline
    subprocess.run(['touch', f'{output_dir}/snakemake/hisat3n'], check=True)
    return



from .hisat3n_general import \
    separate_unique_and_multi_align_reads, \
    convert_hisat_bam_strandness, \
    make_snakefile_hisat3n
from .utilities import validate_cwd_fastq_paths, read_mapping_config
from .hisat3n_mct import select_mct_reads_normal, aggregate_feature_counts
from .summary import snmc_summary, snmct_summary



import pathlib

import pandas as pd
import pysam

from .mc import mc_mapping_stats
from .mct import _count_reads_by_rg_in_star_bam, \
    summary_rna_mapping, \
    summarize_select_dna_reads, \
    aggregate_feature_counts
from .m3c import m3c_mapping_stats


def _4m_mapping_stats(output_dir, config):
    """this may apply to single UID dir, so config is provided as parameter"""
    m3c_stats_df = m3c_mapping_stats(output_dir, config)
    select_dna_stats_df = summarize_select_dna_reads(output_dir, config)
    rna_stats_df = summary_rna_mapping(output_dir)
    final_df = pd.concat([m3c_stats_df, select_dna_stats_df, rna_stats_df], axis=1)
    return final_df


def _4m_additional_cols(final_df, output_dir):
    final_df = final_df.copy()
    final_df['CellInputReadPairs'] = final_df['R1InputReads'].astype(int)
    if 'PCRIndex' in final_df.columns:  # plate info might not exist if the cell name is abnormal
        cell_barcode_ratio = pd.concat([(i['CellInputReadPairs'] / i['CellInputReadPairs'].sum())
                                        for _, i in final_df.groupby('PCRIndex')])
        final_df['CellBarcodeRatio'] = cell_barcode_ratio

    # snm3C part
    final_df['FinalmCReads'] = final_df['R1DeduppedReads'] + final_df['R2DeduppedReads']
    # use % to be consistent with others
    final_df['R1MappingRate'] = final_df['R1UniqueMappedReads'] / final_df['R1TrimmedReads'] * 100
    final_df['R2MappingRate'] = final_df['R2UniqueMappedReads'] / final_df['R2TrimmedReads'] * 100
    final_df['R1DuplicationRate'] = (1 - final_df['R1DeduppedReads'] / final_df['R1UniqueMappedReads']) * 100
    final_df['R2DuplicationRate'] = (1 - final_df['R2DeduppedReads'] / final_df['R2UniqueMappedReads']) * 100
    final_df['TotalContacts'] = final_df[
        ['CisShortContact', 'CisLongContact', 'TransContact']].sum(axis=1)
    final_df['CisShortRatio'] = final_df['CisShortContact'] / final_df['TotalContacts']
    final_df['CisLongRatio'] = final_df['CisLongContact'] / final_df['TotalContacts']
    final_df['TransRatio'] = final_df['TransContact'] / final_df['TotalContacts']

    # snmCT part
    stats = pd.read_hdf(output_dir / 'TotalRNAData.h5', key='stats')
    final_df['GenesDetected'] = stats['GenesDetected']
    # calculate some mCT specific ratios
    final_df['DNAReadsYield'] = final_df['FinalDNAReads'] / (
            final_df['CellInputReadPairs'] * 2)
    final_df['RNAReadsYield'] = final_df['FinalRNAReads'] / final_df[
        'CellInputReadPairs']
    final_df['RNA/(DNA+RNA)'] = final_df['FinalRNAReads'].fillna(0) / (
            final_df['R1DeduppedReads'].fillna(0) + 1)
    return final_df



"""
Each cell_parser function takes a path from single cell and return a series named by cell_id

"""

import pathlib

import numpy as np
import pandas as pd

from .stats_col_names import COL_NAMES


def cell_parser_hisat_summary(stat_path):
    """
    parse hisat3n summary file
    """
    cell_id = pathlib.Path(stat_path).name.split('.')[0]
    term_dict = {
        'Total pairs': 'ReadPairsMappedInPE',
        'Aligned concordantly or discordantly 0 time': 'PEUnmappableReadPairs',
        'Aligned concordantly 1 time': 'PEUniqueMappedReadPairs',
        'Aligned concordantly >1 times': 'PEMultiMappedReadPairs',
        'Aligned discordantly 1 time': 'PEDiscordantlyUniqueMappedReadPairs',
        'Total unpaired reads': 'ReadsMappedInSE',
        'Aligned 0 time': 'SEUnmappableReads',
        'Aligned 1 time': 'SEUniqueMappedReads',
        'Aligned >1 times': 'SEMultiMappedReads',
    }

    with open(stat_path) as rep:
        report_dict = {}
        for line in rep:
            try:
                start, rest = line.split(':')
                start = start.strip()
            except ValueError:
                continue  # more or less than 2 after split
            try:
                report_dict[term_dict[start]] = rest.strip().split(
                    ' ')[0].strip('%')
            except KeyError:
                pass
        for v in term_dict.values():
            if v not in report_dict:
                report_dict[v] = 0

        report_dict = pd.Series(report_dict).astype(int)
        total_reads = report_dict[f'ReadPairsMappedInPE'] * 2 + report_dict[
            'ReadsMappedInSE']
        unique_mapped_reads = report_dict[f'PEUniqueMappedReadPairs'] * 2 + \
                              report_dict[f'PEDiscordantlyUniqueMappedReadPairs'] * 2 + \
                              report_dict[f'SEUniqueMappedReads']
        report_dict['UniqueMappedReads'] = unique_mapped_reads
        report_dict[f'UniqueMappingRate'] = round(unique_mapped_reads /
                                                  (total_reads + 0.00001) * 100)
        multi_mapped_reads = report_dict[f'PEMultiMappedReadPairs'] * 2 + \
                             report_dict[f'SEMultiMappedReads']
        report_dict['MultiMappedReads'] = multi_mapped_reads
        report_dict[f'MultiMappingRate'] = round(multi_mapped_reads /
                                                 (total_reads + 0.00001) * 100)
        report_dict[f'OverallMappingRate'] = round(
            (unique_mapped_reads + multi_mapped_reads) / (total_reads + 0.00001) * 100)
    return pd.Series(report_dict, name=cell_id)


def cell_parser_picard_dedup_stat(stat_path):
    stat_path = pathlib.Path(stat_path)
    cell_id, *other_parts = stat_path.name.split('.')
    try:
        record = pd.read_csv(stat_path, comment='#', sep='\t').T[0]
        record['FinalReads'] = int(record['UNPAIRED_READS_EXAMINED']) - \
                               int(record['UNPAIRED_READ_DUPLICATES']) + \
                               (int(record['READ_PAIRS_EXAMINED']) - int(record['READ_PAIR_DUPLICATES'])) * 2
        record['DuplicatedReads'] = int(record['UNPAIRED_READ_DUPLICATES']) + \
                                    int(record['READ_PAIR_DUPLICATES']) * 2
        record['PCRDuplicationRate'] = record['FinalReads'] / \
                                       (record['FinalReads'] + record['DuplicatedReads'])
        record['PCRDuplicationRate'] = int(
            (1 - record['PCRDuplicationRate']) * 100)

        record.name = cell_id
    except pd.errors.EmptyDataError:
        # if a BAM file is empty, picard matrix is also empty
        record = pd.Series(
            {
                k: 0
                for p, k in COL_NAMES.keys()
                if p == 'parse_picard_dedup_stat'
            },
            dtype='int',
            name=cell_id)
    return record


def cell_parser_cutadapt_trim_stats(path):
    path = pathlib.Path(path)

    cell_id = path.name.split('.')[0]
    cell_records = pd.read_csv(path, sep='\t').T.squeeze()
    cell_records.name = cell_id
    return cell_records


def cell_parser_allc_count(path):
    path = pathlib.Path(path)
    cell_id = path.name.split('.')[0]

    allc_counts = pd.read_csv(path, index_col=0)

    if allc_counts.empty:
        return pd.Series([], dtype='O', name=cell_id)

    # remove contexts that contain "N"
    allc_counts = allc_counts.loc[allc_counts.index.map(
        lambda a: 'N' not in a)].copy()

    # find out which position is the C
    try:
        assert allc_counts.index.map(lambda a: len(a)).unique().size == 1
        c_pos = None
        for i in range(0, len(allc_counts.index[0])):
            if allc_counts.index.str[i].unique().size == 1:
                c_pos = i
        assert c_pos is not None
        assert c_pos != len(allc_counts.index[0])
    except AssertionError:
        raise AssertionError(f'Do not understand the mC context in {path}')

    # get mC context
    mc_context = pd.Series(allc_counts.index.str[c_pos + 1] == 'G').map({
        True: 'mCG', False: 'mCH'
    })
    if c_pos > 0:
        # NOMe
        nome_sites = pd.Series(allc_counts.index.str[c_pos - 1] == 'G').map({
            True: 'G', False: 'H'
        })
        mc_context = nome_sites + mc_context
    mc_context.index = allc_counts.index

    # generate cell records
    mc_context_sum = allc_counts.groupby(mc_context).sum()[['mc', 'cov']]

    is_ccc = allc_counts.index.str[c_pos:].map(
        lambda a: len(set(a)) == 1).values
    if c_pos > 0:
        is_ccc &= np.array(allc_counts.index.str[c_pos - 1] != 'G')

    try:
        ccc_mc, ccc_cov = np.ravel(allc_counts.loc[is_ccc, ['mc', 'cov']].values)
    except ValueError:
        ccc_mc, ccc_cov = 0, 0

    mc_context_sum = pd.concat([
        pd.DataFrame({
            'mCCC': {
                'mc': ccc_mc,
                'cov': ccc_cov
            }
        }).T, mc_context_sum
    ])
    mc_context_sum = mc_context_sum.astype('O')
    mc_context_sum['Frac'] = mc_context_sum['mc'] / (mc_context_sum['cov'] +
                                                     0.00001)
    mc_context_sum.rename(columns={'mc': 'mC', 'cov': 'Cov'}, inplace=True)

    cell_records = {}
    for (count_type, mc_type), count in mc_context_sum.unstack().items():
        cell_records[f'{mc_type}{count_type}'] = count
    cell_records = pd.Series(cell_records, name=cell_id, dtype='O')
    return cell_records


def cell_parser_reads_mc_frac_profile(path):
    # read parameters
    params = {}
    with open(path) as f:
        for line in f:
            if line.startswith('#'):
                k, v = line.strip().split('=')
                params[k[1:]] = v
            else:
                break

    path = pathlib.Path(path)
    cell_id = path.name.split('.')[0]
    reads_profile = pd.read_csv(path, comment='#')
    mode = params['mode'].upper()
    if mode == 'DNA':
        selected_reads = reads_profile[
            (reads_profile['cov'] >= int(params['cov_min_threshold']))
            & (reads_profile['mc_frac'] < float(params['mc_rate_max_threshold']))]
    else:
        selected_reads = reads_profile[
            (reads_profile['cov'] >= int(params['cov_min_threshold']))
            & (reads_profile['mc_frac'] > float(params['mc_rate_min_threshold']))]

    selected_reads = selected_reads['count'].sum()
    selected_ratio = selected_reads / (reads_profile['count'].sum() + 0.0001)
    final_stat = pd.Series({f'Final{mode}Reads': selected_reads,
                            f'Selected{mode}ReadsRatio': selected_ratio},
                           dtype='O', name=cell_id)
    return final_stat


def cell_parser_feature_count_summary(path):
    result = pd.read_csv(path, sep='\t', index_col=0).squeeze()
    all_reads = result.sum()

    result.index.name = None
    result.name = result.name.split(':')[-1]  # cell id
    result['Unassigned_Total'] = result[result.index.str.startswith('Unassigned')].sum()
    result['AssignedRNAReadsRate'] = int(result['Assigned'] / (all_reads + 0.00001) * 100)
    return result


def parse_single_stats_set(path_pattern, parser, prefix=''):
    """
    Parse all the stats files in the path_pattern and return a dataframe
    with each cell id as index and the stats as columns.

    Parameters
    ----------
    path_pattern :
        Path pattern to the stats files.
    parser :
        A function that takes a path and returns a pandas series for one cell.
    prefix :
        Prefix to add to the column names.

    Returns
    -------
    pd.DataFrame
    """
    detail_stats_dir = pathlib.Path('detail_stats/')
    detail_stats_dir.mkdir(exist_ok=True)

    stats_paths = list(pathlib.Path().glob(path_pattern))

    records = []
    empty_index = []
    for path in stats_paths:
        try:
            record = parser(path)
            if record.size == 0:
                empty_index.append(record.name)
            else:
                records.append(record)
        except BaseException as e:
            print(f'Got error when reading {path} with {parser}')
            raise e
    stats_df = pd.DataFrame(records)
    if len(empty_index) > 0:
        use_index = pd.Index(stats_df.index.tolist() + empty_index)
        stats_df = stats_df.reindex(use_index)  # still record empty entries, but values will be nan

    # before rename, save raw stats into detail_stats/
    stats_df.to_csv(f'detail_stats/{prefix}.{parser.__name__}.csv')

    # rename columns
    rename_dict = {}
    rename_functions = set([k[0] for k in COL_NAMES.keys()])
    if parser.__name__ not in rename_functions:
        # do not rename function output if the function name not in COL_NAMES
        return stats_df

    for (parser_name, original_name), new_name in COL_NAMES.items():
        if parser_name == parser.__name__:
            if new_name == '':
                rename_dict[original_name] = original_name
            else:
                rename_dict[original_name] = new_name
    new_columns = stats_df.columns.map(rename_dict)
    if new_columns.isna().sum() != 0:
        print(
            f'These columns are ignored because they do not listed'
            f' in the COL_NAMES dict: {stats_df.columns[new_columns.isna()]}')
    # remove columns marked as DELETE in COL_NAMES
    stats_df.columns = new_columns
    stats_df = stats_df.iloc[:, new_columns != 'DELETE'].copy()

    # add prefix to stats_df columns
    stats_df.columns = stats_df.columns.map(lambda a: prefix + str(a))
    return stats_df



import dnaio
import re

R1_CUT_SITES = {
    'CATG',  # NlaIII
    'CATA',  # NlaIII
    'GATC',  # DpnII or MboI
    'AATC'  # DpnII or MboI
}
R2_CUT_SITES = {
    'CATG',  # NlaIII
    'TATG',  # NlaIII
    'GATC',  # DpnII or MboI
    'GATT'  # DpnII or MboI
}


def _span_combination(spans, min_length):
    """Turn spans into a list of slice object,
    spans is a list such as [(3, 7), (7, 11), (68, 72), (144, 148)]"""
    slices = []
    if len(spans) == 0:
        raise
    elif len(spans) == 1:
        base_start, base_end = spans[0]
        slices.append(slice(base_start, base_end))
    else:
        n_span = len(spans)
        for length in range(1, n_span + 1):
            for start in range(0, n_span - length):
                base_start = spans[start][0]  # start of start
                base_end = spans[start + length][1]  # end of end
                if base_end - base_start >= min_length:
                    # remove small slices
                    slices.append(slice(base_start, base_end))
    return slices


def _make_split_pattern(read_type):
    if read_type.upper() == 'R1':
        _alt_combine = '|'.join(R1_CUT_SITES)
        # like r'(CATG|GATC|CATA|AATC)'
        split_pattern = re.compile(f'({_alt_combine})')
    elif read_type.upper() == 'R2':
        _alt_combine = '|'.join(R2_CUT_SITES)
        split_pattern = re.compile(f'({_alt_combine})')
    else:
        raise ValueError(f'read_type must be R1 or R2, got {read_type}')
    return split_pattern


def _split_read_and_make_combination(read, split_pattern, min_length=30):
    # search enzyme cut sites
    read_length = len(read.sequence)
    spans = [i.span() for i in re.finditer(split_pattern, read.sequence)]
    if len(spans) == 0:
        # if there is not cutsite
        spans = [(0, 0), (read_length, read_length)]
    else:
        if spans[0][0] != 0:
            # if the first span not start at 0, add (0, 0) to capture from the start
            spans = [(0, 0)] + spans
        if spans[-1][1] != read_length:
            # if the last span not end at read_length, add (read_length, read_length) to capture to the end
            spans.append([read_length, read_length])
    # example spans: [(0, 0), (3, 7), (7, 11), (68, 72), (132, 132)]

    # split reads
    slices = _span_combination(spans, min_length=min_length)
    for _slice in slices:
        _read = read[_slice]
        yield _read, _slice


def _trim_site(read_and_slice, read_type):
    """Remove DpnII or MboI site from the left read,
    remove NalIII site from the right read"""
    read, slice = read_and_slice
    start = slice.start
    stop = slice.stop

    sequence = read.sequence
    if read_type == 'R1':
        if sequence[-3:] == 'ATC':
            # If 3' is DpnII or MboI site, clip it (from the left read)
            read = read[:-4]
            stop -= 4
        if sequence[:3] == 'CAT':
            # If 5' is NalIII site, clip it (from the right read)
            read = read[4:]
            start += 4
    else:
        if sequence[-4:-1] == 'GAT':
            # If 3' is DpnII or MboI site, clip it (from the left read)
            read = read[:-4]
            stop -= 4
        if sequence[1:4] == 'ATG':
            # If 5' is NalIII site, clip it (from the right read)
            read = read[4:]
            start += 4
    read.name += f':{start}:{stop}'
    return read, (start, stop)


def split_reads(fastq_path, output_path, read_type, min_length=30):
    """
    Split trimmed fastq file by all possible enzyme cut sites and save to a new fastq file
    Read name is modified, last two field separated by ":" indicate the span start and stop

    Parameters
    ----------
    fastq_path
        Input fastq path
    output_path
        Output fastq path
    read_type
        R1 or R2
    min_length
        Minimum read length to keep

    Returns
    -------

    """
    # determine split pattern
    split_pattern = _make_split_pattern(read_type)
    with dnaio.open(fastq_path) as f, dnaio.open(output_path,
                                                 mode='w') as out_f:
        for read in f:
            read_split_iter = _split_read_and_make_combination(
                read=read, split_pattern=split_pattern, min_length=min_length)
            range_set = set()
            for i, read_and_slice in enumerate(read_split_iter):
                # remove overlapping enzyme cut site
                read_split, range = _trim_site(read_and_slice, read_type)
                if range in range_set:
                    # in some rare cases, the cut site is next to each other,
                    # trim site might cause duplicated ranges,
                    # which is not allowed in aligners
                    continue
                else:
                    range_set.add(range)
                    out_f.write(read_split)
    return
